package ch.ethz.karto.map3d;

import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import javax.swing.event.MouseInputAdapter;

/**
 * Mouse event support for a Map3DViewer. Rotate and zoom for a 3D model.
 * @author Benrhard Jenny, Institute of Cartography, ETH Zurich.
 */
public final class Map3DMouseHandler extends MouseInputAdapter implements MouseWheelListener {

    /**
     * Divide the horizontal mouse movemement by this factor to convert to a
     * rotation angle around the z axis.
     */
    private static final float Z_MOVEMENT_SCALE = 20;
    /**
     * Divide the vertical mouse movemement by this factor to convert to a
     * rotation angle around the x axis.
     */
    private static final float X_MOVEMENT_SCALE = 10;
    /**
     * Divide the number of ticks the scroll wheel was turned by this factor to
     * convert to the zoom factor of the map.
     */
    private static final float SCROLL_WHEEL_SCALE = 10;
    /**
     * Remembers the last position of the mouse.
     */
    private Point lastPoint = null;
    /**
     * Rotate and zoom this map.
     */
    private Map3DViewer map3DViewer;

    /**
     * Construct a new Map3DMouseHandler.
     * @param map3DViewer The map to rotate and zoom.
     */
    public Map3DMouseHandler(Map3DViewer map3DViewer) {
        this.map3DViewer = map3DViewer;
    }

    public boolean isDragging() {
        return lastPoint != null;
    }

    /**
     * Rotate around the vertical z axis and around the horizontal x axis.
     * @param e
     */
    @Override
    public void mouseDragged(MouseEvent e) {
        if (lastPoint == null || map3DViewer == null) {
            return;
        }

        if (e.isShiftDown()) {
            pan(e);
        } else {
            if (this.map3DViewer.is2D()) {
                rotate2D(e);
            } else {
                rotate3D(e);
            }
        }
        this.lastPoint = e.getPoint();
    }

    private void rotate2D(MouseEvent e) {
        int w = this.map3DViewer.getComponent().getWidth();
        int h = this.map3DViewer.getComponent().getHeight();
        float zAngle = this.map3DViewer.getZAngle();
        double lastAngle = Math.atan2(lastPoint.y - h / 2, lastPoint.x - w / 2);
        double newAngle = Math.atan2(e.getY() - h / 2, e.getX() - w / 2);
        zAngle += (float) Math.toDegrees(newAngle - lastAngle);
        this.map3DViewer.setZAngle(zAngle);
    }

    private void pan(MouseEvent e) {
        int h = map3DViewer.getComponent().getWidth();
        float dx = (float) (lastPoint.x - e.getX()) / h * map3DViewer.viewDistance;
        float dy = (float) (lastPoint.y - e.getY()) / h * map3DViewer.viewDistance;
        map3DViewer.setShift(map3DViewer.getShiftX() - dx, map3DViewer.getShiftY() - dy);
/*
        Map3DAnimation animation = map3DViewer.getAnimation();
        Map3DShiftAnimation shiftAnimation;
        if (animation instanceof Map3DShiftAnimation) {
            shiftAnimation = (Map3DShiftAnimation) animation;
        } else {
            shiftAnimation = new Map3DShiftAnimation(map3DViewer, (GLAutoDrawable) map3DViewer.getComponent());
            map3DViewer.setAnimation(shiftAnimation);
            shiftAnimation.startAnimation();
        }

        float targetShiftX = shiftAnimation.targetShiftX - dx;
        float targetShiftY = shiftAnimation.targetShiftY - dy;
        shiftAnimation.animateShift(targetShiftX, targetShiftY);

        ((GLAutoDrawable) map3DViewer.getComponent()).display();
*/    }

    private void rotate3D(MouseEvent e) {
        float zAngle = this.map3DViewer.getZAngle();
        float xAngle = this.map3DViewer.getXAngle();
        zAngle += (lastPoint.x - e.getX()) / Z_MOVEMENT_SCALE;
        xAngle += (lastPoint.y - e.getY()) / X_MOVEMENT_SCALE;
        this.map3DViewer.setZAngle(zAngle);
        this.map3DViewer.setXAngle(xAngle);
    }

    /**
     * Remember the point where the dragging starts.
     * @param e
     */
    @Override
    public void mousePressed(MouseEvent e) {
        this.lastPoint = e.getPoint();
    }

    /**
     * Release the last position of the mouse.
     * @param e
     */
    @Override
    public void mouseReleased(MouseEvent e) {
        this.lastPoint = null;

        // redraw with antialiasing enabled
        if (map3DViewer.isAntialiasing())
            this.map3DViewer.display();
    }

    /**
     * Zoom in and out.
     * @param e
     */
    @Override
    public void mouseWheelMoved(MouseWheelEvent e) {
        float viewDist = this.map3DViewer.getViewDistance();
        viewDist += e.getWheelRotation() / SCROLL_WHEEL_SCALE;
        this.map3DViewer.setViewDistance(viewDist);
    }
}