package ch.ethz.karto.map3d;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import javax.media.opengl.GL;
import javax.media.opengl.GLException;

/**
 * Vertex and fragment shader programs.
 * @author Bernhard Jenny, Institute of Cartography, ETH ZUrich
 */
public class Map3DShader {
    
    private int vertexShader, fragmentShader, shaderProgram;
    private String vShaderCode, fShaderCode;

    protected Map3DShader() {
    }

    /**
     * Compiles and links a fragment and a vertex shader and creates and links
     * a shader program and attaches the shaders. At the end the shaders are
     * ready to be used.
     * @param gl
     * @param vShaderCode The vertex shader code.
     * @param fShaderCode The fragment shader code.
     * @param attribName Array of attribute names used in shaders. Can be null.
     * @param attribID Array of indices associated with the attribute names.
     * Can be null.
     */
    protected void enableShaders(GL gl, String vShaderCode, String fShaderCode,
            String[] attribName, int[] attribID) {

        
        this.vShaderCode = vShaderCode;
        this.fShaderCode = fShaderCode;
        enableShaders(gl, attribName, attribID);
    }

    /**
     * Compiles and links a fragment and a vertex shader and creates and links
     * a shader program and attaches the shaders. At the end the shaders are
     * ready to be used.
     * @param gl
     * @param attribName Array of attribute names used in shaders. Can be null.
     * @param attribID Array of indices associated with the attribute names.
     * Can be null.
     */
    protected void enableShaders(GL gl, String[] attribName, int[] attribID) {

        // crate a shader program
        shaderProgram = gl.glCreateProgram();
        if (shaderProgram == 0) {
            throw new GLException("Shader program could not be created.");
        }

        // attach a vertex shader
        vertexShader = compileShader(gl, vShaderCode, GL.GL_VERTEX_SHADER);
        gl.glAttachShader(shaderProgram, vertexShader);

        // attach a fragment shader
        fragmentShader = compileShader(gl, fShaderCode, GL.GL_FRAGMENT_SHADER);
        gl.glAttachShader(shaderProgram, fragmentShader);

        // bind attribute names used in shaders to predefined locations
        // slot 0 should be used, normally for vertices
        // http://www.gamedev.net/community/forums/topic.asp?topic_id=413008
        if (attribName != null && attribID != null) {
            assert attribName.length == attribID.length;
            for (int i = 0; i < attribID.length; i++) {
                gl.glBindAttribLocation(shaderProgram, attribID[i], attribName[i]);
            }
        }

        // link the program
        gl.glLinkProgram(shaderProgram);

        // check for linking error
        int[] params = new int[1];
        gl.glGetProgramiv(shaderProgram, GL.GL_LINK_STATUS, params, 0);
        if (params[0] == 0) {
            // failure to link
            gl.glGetProgramiv(shaderProgram, GL.GL_LINK_STATUS, params, 0);
            String msg = "Error on shader program linking";
            if (params[0] > 0) {
                byte[] log = new byte[params[0]];
                gl.glGetProgramInfoLog(shaderProgram, params[0], params, 0, log, 0);
                msg += ": " + new String(log, 0, log.length - 1);
                gl.glDeleteProgram(shaderProgram);
            }
            throw new GLException(msg);
        }

        gl.glUseProgram(shaderProgram);
    }

    protected void useProgram(GL gl) {
        int[] i = new int[1];
        gl.glGetIntegerv(GL.GL_CURRENT_PROGRAM, i, 0);
        if (i[0] != shaderProgram) {
            gl.glUseProgram(shaderProgram);
        }
    }

    protected int getAttribLocation(GL gl, String attribName) {
        return gl.glGetAttribLocation(shaderProgram, attribName);
    }

    /**
     * Validate a shader program before rendering. This should only be called
     * during application development, because it can severly hinder performance.
     * @param gl
     * @return
     */
    protected boolean validateProgram(GL gl) {

        // generate information log
        gl.glValidateProgram(shaderProgram);
        int err = gl.glGetError();
        switch (err) {
            case GL.GL_NO_ERROR:
                break;
            case GL.GL_INVALID_VALUE:
                System.err.println("Error on shader program validation: program is not a value generated by OpenGL.");
                return false;
            case GL.GL_INVALID_OPERATION:
                System.err.println("Error on shader program validation: program is not a program object.");
                return false;
            default:
                System.err.println("Error on shader program validation: unknown error.");
                return false;
        }
        
        // retreive information log
        int[] params = new int[1];
        gl.glGetProgramiv(shaderProgram, GL.GL_VALIDATE_STATUS, params, 0);

        if (params[0] == 0) {
            // validation failed
            gl.glGetProgramiv(shaderProgram, GL.GL_INFO_LOG_LENGTH/*GL_LINK_STATUS*/, params, 0);
            String msg = "Error on shader program validation";
            if (params[0] > 0) {
                byte[] log = new byte[params[0]];
                gl.glGetProgramInfoLog(shaderProgram, params[0], params, 0, log, 0);
                msg += ": " + new String(log, 0, params[0]);
                gl.glDeleteProgram(shaderProgram);
            }
            System.err.println(msg);
        }
        return params[0] != 0;
    }

    protected void disableShaders(GL gl) {

        gl.glUseProgram(0);

        // glDeleteProgram and glDeleteShader silently ignore a value of 0,
        // but invalid IDs generate an OpenGL error
        if (gl.glIsProgram(shaderProgram)) {
            gl.glDeleteProgram(shaderProgram);
        }
        shaderProgram = 0;

        if (gl.glIsShader(vertexShader)) {
            gl.glDeleteShader(vertexShader);
        }
        vertexShader = 0;

        if (gl.glIsShader(fragmentShader)) {
            gl.glDeleteShader(fragmentShader);
        }
        fragmentShader = 0;
    }

    private static int compileShader(GL gl, String src, int shaderType) {

        int[] params = new int[1];
        int shaderID = gl.glCreateShader(shaderType);

        if (shaderID != 0 && src != null) {
            gl.glShaderSource(shaderID, 1, new String[]{src}, (int[]) null, 0);
            gl.glCompileShader(shaderID);

            // check for compilation errors
            gl.glGetShaderiv(shaderID, GL.GL_COMPILE_STATUS, params, 0);
            if (params[0] == 0) {
                int[] buffer = new int[1];
                gl.glGetShaderiv(shaderID, GL.GL_INFO_LOG_LENGTH, buffer, 0);
                String msg = "Error on shader compilation";
                if (buffer[0] > 0) {
                    byte[] log = new byte[buffer[0]];
                    gl.glGetShaderInfoLog(shaderID, buffer[0], buffer, 0, log, 0);
                    msg += ": " + new String(log, 0, log.length - 1);
                    gl.glDeleteShader(shaderID);
                }
                throw new GLException(msg);
            }
        }
        return shaderID;
    }

    public static String loadShader(String shaderPath) {
        BufferedReader reader = null;
        try {
            java.net.URL url = Map3DViewer.class.getResource(shaderPath);
            BufferedInputStream bis = new BufferedInputStream(url.openStream());
            InputStreamReader isr = new InputStreamReader(bis, "UTF-8");
            reader = new BufferedReader(isr);
            String line, shader = "";
            while ((line = reader.readLine()) != null) {
                shader += line + "\n";
            }
            return shader;
        } catch (IOException ex) {
            ex.printStackTrace();
            return null;
        } finally {
            try {
                if (reader != null) {
                    reader.close();
                }
            } catch (Exception ex) {
                ex.printStackTrace();
                return null;
            }
        }
    }

    protected void setUniform(GL gl, String name, boolean i) {
        int loc = gl.glGetUniformLocation(shaderProgram, name);
        if (loc != -1) {
            // a bool in a vertex shader is set using an int
            gl.glUniform1i(loc, i ? 1 : 0);
        } else {
            throw new GLException("invalid uniform location: " + name);
        }
    }

    protected void setUniform(GL gl, String name, int i) {
        int loc = gl.glGetUniformLocation(shaderProgram, name);
        if (loc != -1) {
            gl.glUniform1i(loc, i);
        } else {
            throw new GLException("invalid uniform location: " + name);
        }
    }

    protected void setUniform(GL gl, String name, float v) {
        int loc = gl.glGetUniformLocation(shaderProgram, name);
        if (loc != -1) {
            gl.glUniform1f(loc, v);
        } else {
            throw new GLException("invalid uniform location: " + name);
        }
    }

    protected void setUniform(GL gl, String name, float v1, float v2, float v3) {
        int loc = gl.glGetUniformLocation(shaderProgram, name);
        if (loc != -1) {
            gl.glUniform3fv(loc, 1, new float[]{v1, v2, v3}, 0);
        } else {
            throw new GLException("invalid uniform location: " + name);
        }
    }

    protected void setUniform(GL gl, String name, float v1, float v2) {
        int loc = gl.glGetUniformLocation(shaderProgram, name);
        if (loc != -1) {
            gl.glUniform2fv(loc, 1, new float[]{v1, v2}, 0);
        } else {
            throw new GLException("invalid uniform location: " + name);
        }
    }
}
