/*
 * ImageFrame.java
 *
 * Created on June 1, 2008, 4:28 PM
 */
package ika.gui;

import ch.ethz.karto.map3d.Map3DNonLinearTexture1DMapper;
import ch.ethz.karto.map3d.Map3DGLCapabilities;
import ch.ethz.karto.map3d.Map3DTexture;
import ch.ethz.karto.map3d.gui.Map3DOptionsPanel;
import ch.ethz.karto.map3d.Map3DViewer;
import ch.ethz.karto.map3d.Map3DViewer.Camera;
import ch.ethz.karto.map3d.Map3DTexture1DMapper;
import ch.ethz.karto.map3d.gui.Map3DLODPanel;
import ch.ethz.karto.swa.atlas.SystemInfoGPU;
import com.fizzysoft.sdu.RecentDocumentsManager;
import ika.geo.GeoGrid;
import ika.geo.GeoImage;
import ika.geo.GeoObject;
import ika.geo.grid.GaussianPyramid;
import ika.geo.grid.GridScaleOperator;
import ika.geo.grid.Interpolator;
import ika.geoexport.ESRIASCIIGridWriter;
import ika.geoimport.ESRIASCIIGridReader;
import ika.geoimport.GeoImporter;
import ika.geoimport.ImageImporter;
import ika.geoimport.SynchroneDataReceiver;
import ika.utils.ErrorDialog;
import ika.utils.FileUtils;
import ika.utils.ImageUtils;
import ika.utils.MathUtils;
import ika.utils.PropertiesLoader;
import ika.utils.Sys;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.net.URISyntaxException;
import java.util.logging.Level;
import java.util.logging.Logger;
import net.roydesign.mac.MRJAdapter;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URI;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.prefs.Preferences;
import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JRadioButton;
import javax.swing.JSlider;

/**
 * Terrain Bender main window.
 * @author  Helen Jenny and Bernhard Jenny, Institute of Cartography, ETH Zurich.
 */
public class TerrainBenderWindow extends javax.swing.JFrame implements
        PropertyChangeListener, ComponentListener {
    public static final int RECOMMENDED_MAX_TEXTURE_SIZE = 5000;

    /**
     * magic key to identify settings files
     */
    private static final String FILE_FORMAT_IDENTIFIER = "Terrain Bender Format 1.0";
    /**
     * smallest side length of the grid used for previewing
     */
    private static final int MIN_GRID_CELLS = 256 * 256;

    /**
     * Maximum terrain value that Natural Scene Designer can handle.
     */
    private static final float NSD_MAX = 32000;

    /**
     * Minimum terrain value that Natural Scene Designer can handle.
     */
    private static final float NSD_MIN = -16000;


    /**
     * maximum side length of a texture image
     */
    private static final int MAX_TEXTURE_SIDE = 2048;

    private boolean verbose = false;

    /**
     * OpenGL handler
     */
    private Map3DViewer map3DViewer;
    /**
     * information string for the full resolution grid
     */
    private String originalInfoText;
    /**
     * A pyramid of grid levels.
     */
    private GaussianPyramid gridPyramid;
    /** dimension of the window when minimized.
     */
    private java.awt.Dimension packedSize = null;
    private boolean updatingGUI = false;

    private RecentDocumentsManager rdm;

    private Map3DLODPanel lodPanel = new Map3DLODPanel();

    /**
     * Model stores all settings necessary to compute a bent grid.
     */
    private class Model {

        public static final float DEFAULT_BENDING_DIRECTION = 0f;
        /**
         * The direction of bending in degrees.
         */
        public float directionAngle = DEFAULT_BENDING_DIRECTION;
        /**
         * The curve defining the bending. Instance shared with the
         * GradationGraph component.
         */
        public CatmullRomSpline curve;
        /**
         * The curve is scaled by this factor to convert to world coordinates.
         */
        public float basisScale = 0;
        /**
         * Scale factor for the background of the grid.
         */
        public float backgroundScale = 1;
        /**
         * Scale factor for the foreground of the grid.
         */
        public float foregroundScale = 1;
        /**
         * Angle of horizon curvature in degrees.
         */
        public int horizonAngle = 0;
        /**
         * Horizontal position of cylindrical camera relative to size of grid [-1,1]
         */
        public float shiftX = 0;
        /**
         * Vertical position of cylindrical camera relative to size of grid [-1,1]
         */
        public float shiftY = 0;
    }
    private Model p = new Model();

    public TerrainBenderWindow() throws IOException {
        assert activateVerbose();
        
        initRecentDocumentsMenu(this);
        initComponents();
        initMenus();

        // get hold of the shared curve instance.
        p.curve = this.gradationGraph.getCurve(0);

        // apply slight bending to the curve
        this.resetBaseCurve();

        // setup 3D map
        this.map3DViewer = new Map3DViewer(Map3DViewer.GLComponentType.GL_AWT);
        this.map3DViewer.setDefaultCamera(75, 0, Map3DViewer.MAX_DISTANCE, 50, 0, 0);
        this.map3DViewer.resetToDefaultCamera();

        // setup GUI controls of 3D map
        this.map3DOptionsPanel.setCameraVisible(Map3DViewer.Camera.orthogonal, false);
        //this.map3DOptionsPanel.setCameraVisible(Map3DViewer.Camera.parallelOblique, false);
        this.map3DOptionsPanel.setCameraVisible(Map3DViewer.Camera.planOblique, false);
        this.map3DOptionsPanel.setCameraVisible(Map3DViewer.Camera.cylindrical, true);
        this.map3DOptionsPanel.setAntialiasingPanelVisible(false);

        // add 3D map to GUI
        Component map3DComponent = map3DViewer.getComponent();
        map3DComponent.setPreferredSize(new Dimension(400, 400));
        this.container3D.add(map3DComponent, BorderLayout.CENTER);
        map3DComponent.addPropertyChangeListener(this);
        this.map3DOptionsPanel.setMap3DViewer(this.map3DViewer);

        this.map3DOptionsPanel.addPanel("Level of Detail", lodPanel);

        // pack the window (= bring it to its smallest possible size)
        // and store this minimum size
        pack();
        packedSize = getSize();

        // register this as a ComponentListener to get resize events.
        // the resize-event-handler makes sure the window is not gettting too small.
        this.addComponentListener(this);
        
    }

    private boolean activateVerbose() {
        verbose = true;
        return verbose;
    }
    
    private void initRecentDocumentsMenu(final Component parent) {
        rdm = new RecentDocumentsManager() {

            private Preferences getPreferences() {
                return Preferences.userNodeForPackage(TerrainBenderWindow.class);
            }

            @Override
            protected byte[] readRecentDocs() {
                return getPreferences().getByteArray("RecentDocuments", null);
            }

            @Override
            protected void writeRecentDocs(byte[] data) {
                getPreferences().putByteArray("RecentDocuments", data);
            }

            @Override
            protected void openFile(File file, ActionEvent event) {
                try {
                    if (file != null) {
                        openDEM(file.getCanonicalPath());
                    }
                } catch (IOException ex) {
                    String msg = "Could not open the terrain model.";
                    String title = "Terrain Bender Error";
                    ErrorDialog.showErrorDialog(msg, title, ex, parent);
                }
            }
        };
    }

    private GeoGrid getFullResolutionGrid() {
        if (gridPyramid == null) {
            return null;
        }
        return gridPyramid.getFullResolutionLevel();
    }

    private void openDEM(final String filePath) {

        if (!ESRIASCIIGridReader.canRead(filePath)) {
            String msg = "The selected file cannot be read.";
            String title = "Error";
            ErrorDialog.showErrorDialog(msg, title, null, this);
            return;
        }

        final boolean wasGeographic = isGeographicCoordinateSystem();

        // release previous grid to free memory
        gridPyramid = null;

        SwingWorkerWithProgressIndicator worker;
        worker = new SwingWorkerWithProgressIndicator<GaussianPyramid>(
                this, "Terrain Bender Data Import", "", true) {

            @Override
            public void done() {
                try {
                    updatingGUI = true;

                    gridPyramid = get(); // also tests for exceptions

                    originalInfoText = getFullResolutionGrid().toStringWithStatistics("<br>");

                    // compute empirical scale factor for transforming between the curve
                    // in the diagram and world units. Take the larger side of the sourceGrid.
                    double cellSize = getProjectedCellSize(getFullResolutionGrid());
                    int gridDim = Math.max(getFullResolutionGrid().getCols(), getFullResolutionGrid().getRows());
                    p.basisScale = (float) (cellSize * gridDim);

                    // increase vertical exaggeration for large grids in geographical projection
                    double gridSize = gridDim * getFullResolutionGrid().getCellSize();
                    if (isGeographicCoordinateSystem() && gridSize >= 10) {
                        // only automatically change if the user has not set the
                        // exaggeration factors manually.
                        if (scaleForegroundSlider.getValue() == 10
                                && scaleBackgroundSlider.getValue() == 10) {
                            scaleForegroundSlider.setValue(100);
                            scaleBackgroundSlider.setValue(100);
                        }
                    } else if (wasGeographic) {
                        // remove vertical exaggeration if the previous grid
                        // was in geographic coordinates, but the new one is not.
                        scaleForegroundSlider.setValue(10);
                        scaleBackgroundSlider.setValue(10);
                    }

                    // set title of window
                    String name = getFullResolutionGrid().getName();
                    if (name != null && name.trim().length() > 0) {
                        String title = name.trim();
                        if (Sys.isWindows()) {
                            Properties props = PropertiesLoader.loadProperties("ika.app.Application");
                            final String appName = props.getProperty("ApplicationName");
                            title += " - " + appName;
                        }
                        setTitle(title);
                    }

                    ActionListener actionListener = new ActionListener() {

                        @Override
                        public void actionPerformed(ActionEvent e) {
                            if (((JRadioButton) e.getSource()).isSelected()) {
                                modelChanged();
                            }
                        }
                    };
                    lodPanel.initLevelOfDetailGUI(gridPyramid, actionListener);
                    rdm.addDocument(new File(filePath), null);

                    updatingGUI = false;
                    modelChanged();
                } catch (Exception ex) {
                    String exmsg = ex.getMessage();
                    if (exmsg != null && exmsg.contains("user canceled")) {
                        return;
                    }
                    //ex.printStackTrace();

                    String msg = "An error occured";
                    String title = "Terrain Bender Error";
                    ErrorDialog.showErrorDialog(msg, title, ex, TerrainBenderWindow.this);
                    return;
                } finally {
                    updatingGUI = false;
                    this.complete();
                }
            }

            @Override
            protected GaussianPyramid doInBackground() throws Exception {

                this.start();

                // read sourceGrid from file
                GeoGrid newGrid = ESRIASCIIGridReader.read(filePath, this);
                if (this.isAborted()) {
                    throw new IllegalStateException("user canceled");
                }

                this.setIndeterminate(true);
                this.setMessage("Downsampling terrain model for display");
                return new GaussianPyramid(newGrid, 9999, MIN_GRID_CELLS);
            }
        };

        worker.setMaxTimeWithoutDialog(1);
        worker.setMessage("Reading terrain model \"" + FileUtils.getFileName(filePath) + "\"");
        worker.execute();
    }

    /**
     * Load and init GUI with DEM
     */
    public void openDEM() {

        String inputGridPath = FileUtils.askFile(null, "Select an ESRI ASCII Grid", true);
        if (inputGridPath == null) {
            return; // user canceled
        }
        openDEM(inputGridPath);
    }

    private void openTexture(final String filePath) {

        try {
            java.net.URL url = ika.utils.URLUtils.filePathToURL(filePath);
            Dimension dim = ImageImporter.getDimensions(url);

            // test for non power of two support
            if (!Map3DGLCapabilities.hasNonPowerOfTwoTextures()
                    && (!MathUtils.isPower2(dim.width)
                    || !MathUtils.isPower2(dim.height))) {
                String msg = "Your graphics card only supports texture images with\n"
                        + "power of two dimensions (e.g. 1024 x 1024 pixels).\n"
                        + "The selected image has " + dim.width
                        + " x " + dim.height + " pixels\n" 
                        + "and therefore cannot be used.";
                String title = "";
                JOptionPane.showMessageDialog(this, msg, title,
                        JOptionPane.ERROR_MESSAGE, null);
                return;
            }

            // make sure the image is not too large
            int maxSize = Map3DTexture.getMaxTextureSize();
            if (dim.width > maxSize || dim.height > maxSize) {
                String msg = "<html>The selected texture image is too large " +
                        "and cannot be displayed. <br>The maximum texture size is " +
                        maxSize + " x " + maxSize + " pixels.</html>";
                String title = "Large Texture Image";
                JOptionPane.showMessageDialog(this, msg, title,
                        JOptionPane.ERROR_MESSAGE, null);
                return;
            }
        } catch (Exception exc) {
            String msg = "The texture image could not be opened.";
            String title = "Terrain Bender Error";
            ErrorDialog.showErrorDialog(msg, title, exc, this);
            exc.printStackTrace();
            return;
        }

        final JFrame frame = this;
        SwingWorkerWithProgressIndicator worker;
        worker = new SwingWorkerWithProgressIndicator<BufferedImage>(
                this, "Terrain Bender Texture Import", "", true) {

            @Override
            public void done() {
                try {
                    try {
                        BufferedImage t = get(); // tests for exceptions
                        map3DViewer.setTextureImage(t);
                    } catch (Exception ex) {
                        String exmsg = ex.getMessage();
                        if (exmsg != null && exmsg.contains("user canceled")) {
                            return;
                        }

                        String msg = "The texture image could not be opened.";
                        String title = "Terrain Bender Error";
                        ika.utils.ErrorDialog.showErrorDialog(msg, title, ex, frame);
                        return;
                    }
                } finally {
                    modelChanged();
                    this.complete();
                }
            }

            @Override
            protected BufferedImage doInBackground() throws Exception {

                this.setProgress(0);
                ImageImporter importer = new ImageImporter();
                importer.setProgressIndicator(this);
                importer.setOptimizeForDisplay(false);
                java.net.URL url = ika.utils.URLUtils.filePathToURL(filePath);
                SynchroneDataReceiver dataReceiver = new SynchroneDataReceiver();
                dataReceiver.setShowMessageOnError(false);
                importer.read(url, dataReceiver, GeoImporter.SAME_THREAD);

                // test whether the image has been successfully read
                if (dataReceiver.hasReceivedError()) {
                    throw new IOException("Could not read image file at " + filePath);
                }

                // retrieve the image
                GeoImage geoImage = (GeoImage) dataReceiver.getImportedData();
                if (geoImage == null) {
                    return null; // user canceled
                }

                BufferedImage t = geoImage.getBufferedImage();
                if (this.isAborted()) {
                    throw new IllegalStateException("user canceled");
                }

                this.setIndeterminate(true);

                while (t.getWidth() * t.getHeight() > MAX_TEXTURE_SIDE * MAX_TEXTURE_SIDE) {
                    int w = t.getWidth() / 2;
                    int h = t.getHeight() / 2;
                    Object hint = RenderingHints.VALUE_INTERPOLATION_BICUBIC;
                    int imageType = BufferedImage.TYPE_INT_ARGB;
                    t = ImageUtils.getFasterScaledInstance(t, w, h, hint, imageType, false);
                    if (this.isAborted()) {
                        throw new IllegalStateException("user canceled");
                    }
                }

                // JOGL seems to have a bug and throws an error if the image
                // used for texturing does not have an alpha channel
                t = ImageUtils.toARGB(t);

                return t;
            }
        };

        worker.setMaxTimeWithoutDialog(1);
        worker.setMessage("Reading Texture \"" + FileUtils.getFileName(filePath) + "\"");
        worker.execute();
    }

    /** 
     * Mac OS X and Windows specific initialization of the menus
     */
    private void initMenus() {
        if (ika.utils.Sys.isMacOSX()) {

            // remove exit menu item on Mac OS X
            this.fileMenu.remove(this.exitMenuSeparator);
            this.fileMenu.remove(this.exitMenuItem);
            this.fileMenu.validate();

            // remove window info menu item on Mac OS X
            this.menuBar.remove(this.winHelpMenu);

            /*
            // remove preferences menu item on Mac OS X
            this.editMenu.remove(this.preferencesSeparator);
            this.editMenu.remove(this.preferencesMenuItem);
            this.editMenu.validate();
             */

            // setup about command in apple menu
            final JFrame owner = this;
            MRJAdapter.addAboutListener(new ActionListener() {

                @Override
                public void actionPerformed(ActionEvent e) {
                    ika.gui.ProgramInfoPanel.showApplicationInfo(owner);
                }
            });

            // setup quit command in apple menu
            MRJAdapter.addQuitApplicationListener(new ActionListener() {

                @Override
                public void actionPerformed(ActionEvent e) {
                    askToCloseWindowAndExit();
                }
            });
        } else if (ika.utils.Sys.isWindows()) {
            this.menuBar.remove(macHelpMenu);
        }

        this.menuBar.validate();
    }

    private void noTerrainErrorMessage(JFrame frame) {
        String msg = "<html>There is no terrain model loaded.<br>" +
                "Please first open one.";
        String title = "Terrain Bender Error";
        JOptionPane.showMessageDialog(frame, msg, title, JOptionPane.INFORMATION_MESSAGE);
    }

    /**
     * Displays a dialog with information about the original sourceGrid and the
     * current bending.
     */
    public void showGridInfo() {

        try {
            
            if (originalInfoText == null 
                    || lodPanel.getDisplayGrid(gridPyramid) == null) {
                noTerrainErrorMessage(this);
                return;
            }

            StringBuilder sb = new StringBuilder();
            sb.append("<html> <b>");
            sb.append("Original Terrain Model");
            sb.append("</b><br><br>");

            sb.append(this.originalInfoText);

            // stats for bent grid
            if (gridPyramid != null) {
                GeoGridWrapper ggw = new GeoGridWrapper();
                // use reasonably sized grid and not full size grid, which can become
                // too large.
                int pyramidDisplayLevel = lodPanel.getDefaultDisplayLevel(gridPyramid);
                GeoGrid gridToBend = gridPyramid.getLevel(pyramidDisplayLevel);
                ggw.init(gridToBend);
                boolean cylindrical = this.map3DViewer.getCamera() == Camera.cylindrical;
                this.bendGrid(gridToBend, ggw, cylindrical);
                if (ggw.grid != null) {
                    sb.append("<br><br><b><br>");
                    sb.append("Bent Terrain Model");
                    sb.append("</b><br><br>");
                    sb.append("(approximate values)<br>");
                    sb.append(ggw.grid.getStatistics().toString(false, "<br>"));
                    sb.append("</html>");
                }
            }

            String title = "Terrain Model Info";
            JOptionPane.showMessageDialog(this, sb.toString(), title, JOptionPane.INFORMATION_MESSAGE);

        } catch (Exception exc) {
            String msg = "An error occured.";
            String title = "Terrain Bender Error";
            ErrorDialog.showErrorDialog(msg, title, exc, this);
        }
    }

    /**
     * AbstractGrid is a destination for storing bent grids.
     */
    private abstract class AbstractGrid {

        public abstract void init(GeoGrid gridTemplage);

        public abstract void setValue(float v, int col, int row);

        public abstract boolean isThreadable();
    }

    /**
     * Wrapper around a GeoGrid, i.e. the grid is stored in RAM.
     */
    private class GeoGridWrapper extends AbstractGrid {

        public GeoGrid grid;

        public GeoGridWrapper() {
        }

        @Override
        public void init(GeoGrid template) {
            int cols = template.getCols();
            int rows = template.getRows();
            double cellsize = template.getCellSize();
            grid = new GeoGrid(cols, rows, cellsize);
            grid.setWest(template.getWest());
            grid.setNorth(template.getNorth());
        }

        @Override
        public final void setValue(float v, int col, int row) {
            grid.setValue(v, col, row);
        }

        @Override
        public boolean isThreadable() {
            // cover unlikely case that there are more CPUs than rows in grid
            final int cpuCount = Runtime.getRuntime().availableProcessors();
            return grid.getRows() > cpuCount;
        }
    }

    /**
     * Wrapper around an ESRIASCIIGridWriter, i.e. the grid is written to
     * disk.
     */
    private class ESRIASCIIGridWriterWrapper extends AbstractGrid {

        private ESRIASCIIGridWriter writer;
        private PrintWriter printWriter;
        private int cols,  rows;
        private ProgressIndicator progress;
        private float offset;

        public ESRIASCIIGridWriterWrapper(PrintWriter printWriter,
                ProgressIndicator progress,
                float offset) {
            this.printWriter = printWriter;
            this.progress = progress;
            this.offset = offset;
        }

        @Override
        public void init(GeoGrid template) {
            this.cols = template.getCols();
            this.rows = template.getRows();
            double cellsize = template.getCellSize();
            double west = template.getWest();
            double south = template.getNorth() - (rows - 1) * cellsize;

           
            writer = new ESRIASCIIGridWriter(printWriter, cols, rows, 
                    west, south, cellsize, -9999);
            progress.start();
        }

        @Override
        public final void setValue(float v, int col, int row) {

            writer.write(v + offset);
            
            if ((col + 1) % cols == 0) {
                writer.newLine();
                if (progress != null) {
                    progress.progress((row + 1) * 100 / rows);
                }
            }

            if (col % 50 == 0) {
                if (progress.isAborted()) {
                    throw new IllegalStateException("user canceled");
                }
            }
        }

        @Override
        public boolean isThreadable() {
            return false;
        }
    }

    /**
     * Extends AbstractGrid for finding minimum value of grid after bending.
     */
    private class MinMaxFinderGrid extends AbstractGrid {

        private int cols,  rows;
        private ProgressIndicator progress;
        private float minValue = Float.MAX_VALUE;
        private float maxValue = -Float.MAX_VALUE;

        private MinMaxFinderGrid(ProgressIndicator progress) {
            this.progress = progress;
        }

        @Override
        public void init(GeoGrid template) {
            this.cols = template.getCols();
            this.rows = template.getRows();
            if (progress != null) {
                progress.start();
            }
        }

        @Override
        public final void setValue(float v, int col, int row) {

            if (v < minValue) {
                minValue = v;
            }
            if (v > maxValue) {
                maxValue = v;
            }

            if (progress != null) {
                if ((col + 1) % cols == 0) {
                    progress.progress((row + 1) * 100 / rows);
                }

                if (col % 50 == 0) {
                    if (progress.isAborted()) {
                        throw new IllegalStateException("user canceled");
                    }
                }
            }
        }

        @Override
        public boolean isThreadable() {
            return false;
        }

        private float getMinValue() {
            return minValue;
        }

        private float getMaxValue() {
            return maxValue;
        }

    }

    /**
     * Returns whether the grid seems to be in geographical coordinates, i.e.
     * cell size is in degrees instead of meters.
     * @return
     */
    private boolean isGeographicCoordinateSystem() {
        GeoGrid grid = this.getFullResolutionGrid();
        return grid == null ? false : grid.getCellSize() < 0.1;
    }

    /**
     * Returns a cell size. Computes a projected cell size if the grid is not
     * projected.
     * @param resampledGrid
     * @return
     */
    private double getProjectedCellSize(GeoGrid resampledGrid) {

        double downsampledCellSize = resampledGrid.getCellSize();

        // the following test of the cell size must be done with the original
        // sourceGrid. Very large geographic grids that have been resampled to a small
        // size for display might end up with relatively large cell sizes.
        if (isGeographicCoordinateSystem()) {
            double lat = (getFullResolutionGrid().getNorth() + getFullResolutionGrid().getSouth()) / 2;
            lat = Math.toRadians(lat);
            downsampledCellSize = 6371000 * Math.toRadians(downsampledCellSize) * Math.cos(lat);
        }
        return downsampledCellSize;
    }

    /**
     * Returns a bent sourceGrid.
     * Important: this method can be executed on a thread other than
     * the event dispatching thread, i.e. it must not interact with any GUI element.
     * @return
     */
     private void bendGrid(final GeoGrid srcGrid,
            final AbstractGrid destGrid,
            final boolean cylindricalBending) throws InterruptedException {

        
        int threadsCount = Runtime.getRuntime().availableProcessors();
        if (!destGrid.isThreadable()) {
            threadsCount = 1;
        }
        Thread threads[] = new Thread[threadsCount];
        int rowsPerCPU = srcGrid.getRows() / threadsCount;

        InterruptingThreadGroup threadGroup = new InterruptingThreadGroup("bender threads");
        for (int i = 0; i < threads.length; i++) {
            int firstRow = i * rowsPerCPU;
            int lastRow = firstRow + rowsPerCPU;
            if (i == threads.length - 1) {
                lastRow = srcGrid.getRows() - 1;
            }
            
            BenderRunnable bender
                    = new BenderRunnable(srcGrid,
                    destGrid, firstRow, lastRow, cylindricalBending, p.curve.clone());
            threads[i] = new Thread(threadGroup, bender, Integer.toString(i));
            threads[i].start();
        }

        for (int i = 0; i < threads.length; i++) {
            threads[i].join();
        }

    }

    /**
     * Thread group that interrupts all its threads if any throws an uncaught exception.
     */
    public class InterruptingThreadGroup extends ThreadGroup {

        public Throwable e;

        public InterruptingThreadGroup(String name) {
            super(name);
        }

        @Override
        public void uncaughtException(Thread t, Throwable e) {
            this.interrupt();
            synchronized (this) {
                if (!(e instanceof InterruptedException)) {
                    this.e = e;
                }
            }
        }
    }

    /**
     * Runnable that bends a part of a grid
     */
    class BenderRunnable implements Runnable {

        private final GeoGrid srcGrid;
        private final AbstractGrid destGrid;
        private final int firstRow;
        private final int lastRow;
        private final boolean cylindricalBending;
        private final CatmullRomSpline curve;

        protected BenderRunnable(GeoGrid srcGrid,
                AbstractGrid destGrid,
                int firstRow,
                int lastRow,
                boolean cylindricalBending,
                CatmullRomSpline curve) {

            this.srcGrid = srcGrid;
            this.destGrid = destGrid;
            this.firstRow = firstRow;
            this.lastRow = lastRow;
            this.cylindricalBending = cylindricalBending;
            this.curve = curve;

        }

        @Override
        public void run() {

            if (cylindricalBending) {
                bendForCylindricalProjection(srcGrid, destGrid, firstRow, lastRow, curve);
            } else {
                bendForCentralPerspective(srcGrid, destGrid, firstRow, lastRow, curve);
            }

        }
    }

    /**
     * computes a bent sourceGrid for viewing with a perspective camera.
     * Important: this method can be executed on a thread other than
     * the event dispatching thread, i.e. it must not interact with any GUI element.
     * @param sourceGrid
     * @return
     */
    private void bendForCentralPerspective(GeoGrid srcGrid,
            AbstractGrid destGrid,
            int firstRow,
            int lastRow,
            CatmullRomSpline curve) {

        if (srcGrid == null) {
            return;
        }

        final int rows = srcGrid.getRows();
        final int cols = srcGrid.getCols();

        // init interpolation
        Interpolator interpolatorInViewDir = new Interpolator();
        interpolatorInViewDir.setFrontBackAngle(p.directionAngle);
        Interpolator interpolatorPerpToViewDir = new Interpolator();
        interpolatorPerpToViewDir.setFrontBackAngle(p.directionAngle + 90);

        // the relative width of the sourceGrid perpendicular to the bending direction
        double rotWidth = interpolatorPerpToViewDir.w(cols, rows) / Math.max(cols, rows);

        // compute the radius of the sphere onto which the sourceGrid is bent
        final double R = rotWidth / 2. / Math.sin(Math.toRadians(p.horizonAngle) / 2);
        final double Rsq = R * R;

        for (int r = firstRow; r <= lastRow; r++) {
            float [] srcRow = srcGrid.getGrid()[r];
            for (int c = 0; c < cols; c++) {

                final float v = srcRow[c];
                if (Float.isNaN(v)) {
                    destGrid.setValue(Float.NaN, c, r);
                    continue;
                }

                // compute distance along bending direction
                final double t = 1. - interpolatorInViewDir.weight(c, r, cols, rows);

                // get curve value of base
                float curveY = curve.evaluateSimilarToPrevious(t);

                // compute base
                final float base;
                if (p.horizonAngle == 0) {
                    base = curveY * p.basisScale;
                } else {
                    // apply spherical bending

                    // relative distance perpendicular to the direction of bending
                    // with a point on the axis s = 0 and a point at
                    // maximum distance s = 1.
                    final double s = interpolatorPerpToViewDir.weightRelativeToCenter(c, r, cols, rows) * rotWidth / 2;

                    // vertical base at distance s
                    double d = R - Math.sqrt(Rsq - s * s);

                    // linearly decrease bending towards foreground
                    d *= t;

                    // combine base defined by the interactive curve and
                    // the spherical bending
                    base = (float) ((curveY - d) * p.basisScale);
                }

                // compute vertical exaggeration for the terrain
                final double detailsScale =
                        t * p.backgroundScale + (1 - t) * p.foregroundScale;
                
                // add base and vertically exaggerated terrain
                destGrid.setValue(base + v * (float) detailsScale, c, r);
            }
        }
    }

    /**
     * Important: this method can be executed on a thread other than
     * the event dispatching thread, i.e. it must not interact with any GUI element.
     * @param srcGrid
     * @param destGrid
     */
    private void bendForCylindricalProjection(GeoGrid srcGrid,
            AbstractGrid destGrid,
            int firstRow,
            int lastRow,
            CatmullRomSpline curve) {

        if (srcGrid == null) {
            return;
        }

        int rows = srcGrid.getRows();
        int cols = srcGrid.getCols();

        // camera position
        final float camPosX = cols / 2f * (1f + p.shiftX);
        final float camPosY = rows / 2f * (1f + p.shiftY);

        // distance to most distant corner
        double maxD = Math.hypot(camPosX - cols, camPosY - rows);
        maxD = Math.max(maxD, Math.hypot(camPosX, camPosY - rows));
        maxD = Math.max(maxD, Math.hypot(camPosX, camPosY));
        maxD = Math.max(maxD, Math.hypot(camPosX - cols, camPosY));
        final float maxDInv = (float) (1. / maxD);

        for (int r = firstRow; r <= lastRow; r++) {
            final double y = camPosY - r;
            for (int c = 0; c < cols; c++) {

                // relative distance to center of the camera
                final float x = camPosX - c;
                final float d = (float) Math.hypot(x, y);
                final float t = d * maxDInv;

                final float base = curve.evaluateSimilarToPrevious(t) * p.basisScale;
                final float v = srcGrid.getValue(c, r);

                // vertical exaggeration for the terrain
                final float detailsScale =
                        t * p.backgroundScale + (1 - t) * p.foregroundScale;

                destGrid.setValue(base + v * detailsScale, c, r);
            }
        }
    }

    /**
     * Converts to model to a string that can be written to a file for storing.
     * @return
     */
    @Override
    public String toString() {
        DecimalFormat formatter = new DecimalFormat("##0.####");
        DecimalFormatSymbols dfs = formatter.getDecimalFormatSymbols();
        dfs.setDecimalSeparator('.');
        formatter.setDecimalFormatSymbols(dfs);
        String lineSep = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder();

        sb.append(FILE_FORMAT_IDENTIFIER);
        sb.append(lineSep);

        sb.append("Basis Curve");
        sb.append(lineSep);
        sb.append(p.curve.toString());
        sb.append(lineSep);

        sb.append("Direction of View");
        sb.append(lineSep);
        sb.append(formatter.format(p.directionAngle));
        sb.append(lineSep);

        sb.append("Background Exaggeration");
        sb.append(lineSep);
        sb.append(formatter.format(p.backgroundScale));
        sb.append(lineSep);

        sb.append("Foreground Exaggeration");
        sb.append(lineSep);
        sb.append(formatter.format(p.foregroundScale));
        sb.append(lineSep);

        sb.append("Horizon Curvature");
        sb.append(lineSep);
        sb.append(p.horizonAngle);
        sb.append(lineSep);

        sb.append("Horizontal Shift");
        sb.append(lineSep);
        sb.append(p.shiftX);
        sb.append(lineSep);

        sb.append("Vertical Shift");
        sb.append(lineSep);
        sb.append(p.shiftY);
        sb.append(lineSep);

        return sb.toString();
    }

    /**
     * Restores the settings from a string that has been generated by toString
     * @param string
     * @throws java.io.IOException
     */
    public void fromString(String string) throws IOException {

        try {
            updatingGUI = true;

            // make sure we have the right file format
            if (!string.startsWith(FILE_FORMAT_IDENTIFIER)) {
                throw new IOException("Not a Terrain Bender Format 1.0");
            }

            StringTokenizer tokenizer = new StringTokenizer(string, "\n\r");

            // overread Terrain Bender Format 1.0
            tokenizer.nextToken();

            // overread "Basis CatmullRomSpline"
            tokenizer.nextToken();
            p.curve.fromString(tokenizer.nextToken());
            this.gradationGraph.repaint();

            // overread "Direction of View"
            tokenizer.nextToken();
            float dir = Float.parseFloat(tokenizer.nextToken());
            p.directionAngle = dir;
            this.map3DViewer.setZAngle(dir);

            // overread "Background Exaggeration"
            tokenizer.nextToken();
            float v = Float.parseFloat(tokenizer.nextToken()) * 10f;
            scaleBackgroundSlider.setValue((int) v);

            // overread "Foreground Exaggeration"
            tokenizer.nextToken();
            v = Float.parseFloat(tokenizer.nextToken()) * 10f;
            scaleForegroundSlider.setValue((int) v);

            // overread "Horizon Curvature"
            tokenizer.nextToken();
            this.curvatureSlider.setValue(Integer.parseInt(tokenizer.nextToken()));

            // overread "Horizontal Shift"
            tokenizer.nextToken();
            p.shiftX = Float.parseFloat(tokenizer.nextToken());

            // overread "Vertical Shift"
            tokenizer.nextToken();
            p.shiftY = Float.parseFloat(tokenizer.nextToken());
            this.map3DViewer.setShift(p.shiftX, p.shiftY);
        } finally {
            updatingGUI = false;
        }
    }

    
    /**
     * Called whenever the model changes and the bending must be re-generated.
     */
    protected void modelChanged() {

        GeoGrid displayGrid = lodPanel.getDisplayGrid(gridPyramid);
        try {
            if (displayGrid == null || updatingGUI) {
                return;
            }

            this.getRootPane().putClientProperty("Window.documentModified", Boolean.TRUE);

            long modelChangedStartTime = System.currentTimeMillis();

            GeoGridWrapper ggw = new GeoGridWrapper();
            ggw.init(displayGrid);
            boolean cylindrical = this.map3DViewer.getCamera() == Camera.cylindrical;
            bendGrid(displayGrid, ggw, cylindrical);
            long mergingEndTime = System.currentTimeMillis();

            long openGLStartTime = System.currentTimeMillis();
            if (this.map3DViewer != null) {
                double cellSize = getProjectedCellSize(ggw.grid);
                Map3DTexture1DMapper mapper = new Map3DNonLinearTexture1DMapper(displayGrid.getGrid());
                map3DViewer.setModel(ggw.grid.getGrid(), (float) cellSize, mapper);
            }

            if (verbose) {
                long endTime = System.currentTimeMillis();
                if (ggw.grid != null) {
                    System.out.println("Model Changed (" + ggw.grid.getCols() + " x " + ggw.grid.getRows() + ")");
                }
                System.out.println("Bending:  " + (mergingEndTime - modelChangedStartTime) + " ms");
                System.out.println("OpenGL: " + (endTime - openGLStartTime) + " ms");
                System.out.println("Total:  " + (endTime - modelChangedStartTime) + " ms");
                System.out.println();
            }

        } catch (Exception exc) {
            String msg = "An error occured.";
            String title = "Terrain Bender Error";
            ErrorDialog.showErrorDialog(msg, title, exc, this);
        }
    }

    /**
     * Displays a dialog that asks the user whether the application should
     * be closed and does so if the user confirms.
     */
    private void askToCloseWindowAndExit() {
        if (this.getFullResolutionGrid() == null) {
            System.exit(0);
        }

        Map3DOptionsPanel.hide3DOptions();
        int res = JOptionPane.showConfirmDialog(this,
                "<html>This will quit Terrain Bender. <br>Unsaved settings will be lost.</html>",
                "Close Terrain Bender",
                JOptionPane.OK_CANCEL_OPTION);
        if (res == JOptionPane.OK_OPTION) {
            this.dispose();
            System.exit(0);
        }
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {
        java.awt.GridBagConstraints gridBagConstraints;

        scaleTerrainPanel = new javax.swing.JPanel();
        javax.swing.JLabel jLabel8 = new javax.swing.JLabel();
        scaleTerrainFormattedTextField = new javax.swing.JFormattedTextField();
        javax.swing.JLabel jLabel9 = new javax.swing.JLabel();
        cameraPositionPanel = new javax.swing.JPanel();
        javax.swing.JLabel jLabel10 = new javax.swing.JLabel();
        cameraXFormattedTextField = new javax.swing.JFormattedTextField();
        javax.swing.JLabel jLabel11 = new javax.swing.JLabel();
        javax.swing.JLabel jLabel12 = new javax.swing.JLabel();
        cameraYFormattedTextField = new javax.swing.JFormattedTextField();
        javax.swing.JLabel jLabel13 = new javax.swing.JLabel();
        eastLabel = new javax.swing.JLabel();
        jLabel15 = new javax.swing.JLabel();
        jLabel16 = new javax.swing.JLabel();
        southLabel = new javax.swing.JLabel();
        jLabel18 = new javax.swing.JLabel();
        northLabel = new javax.swing.JLabel();
        jLabel20 = new javax.swing.JLabel();
        westLabel = new javax.swing.JLabel();
        container3D = new javax.swing.JPanel();
        eastPanel = new javax.swing.JPanel();
        jPanel1 = new javax.swing.JPanel();
        bendingTabbedPane = new javax.swing.JTabbedPane();
        javax.swing.JPanel settingsPanel = new ika.gui.TransparentMacPanel();
        gradationGraph = new ika.gui.GradationGraph();
        javax.swing.JLabel jLabel2 = new javax.swing.JLabel();
        javax.swing.JLabel jLabel3 = new javax.swing.JLabel();
        javax.swing.JPanel detailsPanel = new ika.gui.TransparentMacPanel();
        javax.swing.JPanel detailsPanel_ = new ika.gui.TransparentMacPanel();
        javax.swing.JLabel jLabel1 = new javax.swing.JLabel();
        javax.swing.JLabel jLabel4 = new javax.swing.JLabel();
        scaleBackgroundSlider = new javax.swing.JSlider();
        scaleForegroundSlider = new javax.swing.JSlider();
        jLabel7 = new javax.swing.JLabel();
        noVerticalExaggerationButton = new javax.swing.JButton();
        noVerticalExaggerationButton.putClientProperty("JButton.buttonType", "roundRect");
        horizonPanel = new ika.gui.TransparentMacPanel();
        javax.swing.JPanel sphericalPanel_ = new ika.gui.TransparentMacPanel();
        javax.swing.JLabel jLabel5 = new javax.swing.JLabel();
        curvatureSlider = new javax.swing.JSlider();
        map3DOptionsPanel = new ch.ethz.karto.map3d.gui.Map3DOptionsPanel();
        directionButton = new javax.swing.JButton();
        menuBar = new javax.swing.JMenuBar();
        fileMenu = new javax.swing.JMenu();
        openDEMMenuItem = new javax.swing.JMenuItem();
        openRecentMenu = rdm.createOpenRecentMenu();
        saveDEMMenuItem = new javax.swing.JMenuItem();
        jSeparator5 = new javax.swing.JSeparator();
        openTextureMenuItem = new javax.swing.JMenuItem();
        clearTextureMenuItem = new javax.swing.JMenuItem();
        jSeparator1 = new javax.swing.JSeparator();
        saveSettingsMenuItem = new javax.swing.JMenuItem();
        openSettingsMenuItem = new javax.swing.JMenuItem();
        jSeparator2 = new javax.swing.JSeparator();
        export3DViewMenuItem = new javax.swing.JMenuItem();
        jMenu1 = new javax.swing.JMenu();
        exportDepthBuffer8BitMenuItem = new javax.swing.JMenuItem();
        exportDepthBuffer16BitMenuItem = new javax.swing.JMenuItem();
        exitMenuSeparator = new javax.swing.JSeparator();
        exitMenuItem = new javax.swing.JMenuItem();
        editMenu = new javax.swing.JMenu();
        cutMenuItem = new javax.swing.JMenuItem();
        copyMenuItem = new javax.swing.JMenuItem();
        pasteMenuItem = new javax.swing.JMenuItem();
        deleteMenuItem = new javax.swing.JMenuItem();
        jSeparator3 = new javax.swing.JSeparator();
        gridInfoMenuItem = new javax.swing.JMenuItem();
        jSeparator4 = new javax.swing.JSeparator();
        bendingDirectionMenuItem = new javax.swing.JMenuItem();
        resetBendingMenuItem = new javax.swing.JMenuItem();
        jSeparator6 = new javax.swing.JSeparator();
        viewingDirectionMenuItem = new javax.swing.JMenuItem();
        jSeparator7 = new javax.swing.JSeparator();
        scaleTerrainMenuItem = new javax.swing.JMenuItem();
        jSeparator8 = new javax.swing.JSeparator();
        cameraPositionMenuItem = new javax.swing.JMenuItem();
        macHelpMenu = new javax.swing.JMenu();
        macHelpMenuItem = new javax.swing.JMenuItem();
        systemInformationMenuItem = new javax.swing.JMenuItem();
        winHelpMenu = new javax.swing.JMenu();
        winHelpMenuItem = new javax.swing.JMenuItem();
        infoMenuItem = new javax.swing.JMenuItem();
        winSystemInformationMenuItem = new javax.swing.JMenuItem();
        extrasMenu = new javax.swing.JMenu();
        highResMenuItem = new javax.swing.JCheckBoxMenuItem();
        bindLightDirectionMenuItem = new javax.swing.JCheckBoxMenuItem();
        updateDisplayMenuItem = new javax.swing.JMenuItem();

        scaleTerrainPanel.setLayout(new java.awt.GridBagLayout());

        jLabel8.setText("Scale Factor:");
        scaleTerrainPanel.add(jLabel8, new java.awt.GridBagConstraints());

        scaleTerrainFormattedTextField.setFormatterFactory(new javax.swing.text.DefaultFormatterFactory(new javax.swing.text.NumberFormatter()));
        scaleTerrainFormattedTextField.setPreferredSize(new java.awt.Dimension(200, 28));
        scaleTerrainFormattedTextField.setValue(new Float(1));
        scaleTerrainPanel.add(scaleTerrainFormattedTextField, new java.awt.GridBagConstraints());

        jLabel9.setFont(jLabel9.getFont().deriveFont(jLabel9.getFont().getSize()-2f));
        jLabel9.setText("All values in the terrain grid are scaled by this factor.");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.gridwidth = 2;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.insets = new java.awt.Insets(10, 0, 0, 0);
        scaleTerrainPanel.add(jLabel9, gridBagConstraints);

        cameraPositionPanel.setLayout(new java.awt.GridBagLayout());

        jLabel10.setText("Horizontal");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        cameraPositionPanel.add(jLabel10, gridBagConstraints);

        cameraXFormattedTextField.setFormatterFactory(new javax.swing.text.DefaultFormatterFactory(new javax.swing.text.NumberFormatter()));
        cameraXFormattedTextField.setPreferredSize(new java.awt.Dimension(200, 28));
        cameraXFormattedTextField.setValue(new Float(1));
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridwidth = 2;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        cameraPositionPanel.add(cameraXFormattedTextField, gridBagConstraints);

        jLabel11.setFont(jLabel11.getFont().deriveFont(jLabel11.getFont().getSize()-2f));
        jLabel11.setText("<html>Position the cylindrical camera. Coordinates are in meters.</html>");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        gridBagConstraints.gridwidth = 3;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.insets = new java.awt.Insets(10, 0, 10, 0);
        cameraPositionPanel.add(jLabel11, gridBagConstraints);

        jLabel12.setText("Vertical");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        cameraPositionPanel.add(jLabel12, gridBagConstraints);

        cameraYFormattedTextField.setFormatterFactory(new javax.swing.text.DefaultFormatterFactory(new javax.swing.text.NumberFormatter()));
        cameraYFormattedTextField.setPreferredSize(new java.awt.Dimension(200, 28));
        cameraYFormattedTextField.setValue(new Float(1));
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.gridwidth = 2;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        cameraPositionPanel.add(cameraYFormattedTextField, gridBagConstraints);

        jLabel13.setFont(jLabel13.getFont().deriveFont(jLabel13.getFont().getSize()-2f));
        jLabel13.setText("Grid extension:");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 3;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        cameraPositionPanel.add(jLabel13, gridBagConstraints);

        eastLabel.setFont(eastLabel.getFont().deriveFont(eastLabel.getFont().getSize()-2f));
        eastLabel.setText("jLabel14");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 2;
        gridBagConstraints.gridy = 4;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 0);
        cameraPositionPanel.add(eastLabel, gridBagConstraints);

        jLabel15.setFont(jLabel15.getFont().deriveFont(jLabel15.getFont().getSize()-2f));
        jLabel15.setText("East");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 4;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 0);
        cameraPositionPanel.add(jLabel15, gridBagConstraints);

        jLabel16.setFont(jLabel16.getFont().deriveFont(jLabel16.getFont().getSize()-2f));
        jLabel16.setText("South");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 5;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 0);
        cameraPositionPanel.add(jLabel16, gridBagConstraints);

        southLabel.setFont(southLabel.getFont().deriveFont(southLabel.getFont().getSize()-2f));
        southLabel.setText("jLabel14");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 2;
        gridBagConstraints.gridy = 5;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 0);
        cameraPositionPanel.add(southLabel, gridBagConstraints);

        jLabel18.setFont(jLabel18.getFont().deriveFont(jLabel18.getFont().getSize()-2f));
        jLabel18.setText("North");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 6;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 0);
        cameraPositionPanel.add(jLabel18, gridBagConstraints);

        northLabel.setFont(northLabel.getFont().deriveFont(northLabel.getFont().getSize()-2f));
        northLabel.setText("jLabel14");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 2;
        gridBagConstraints.gridy = 6;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 0);
        cameraPositionPanel.add(northLabel, gridBagConstraints);

        jLabel20.setFont(jLabel20.getFont().deriveFont(jLabel20.getFont().getSize()-2f));
        jLabel20.setText("West");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 3;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 0);
        cameraPositionPanel.add(jLabel20, gridBagConstraints);

        westLabel.setFont(westLabel.getFont().deriveFont(westLabel.getFont().getSize()-2f));
        westLabel.setText("jLabel14");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 2;
        gridBagConstraints.gridy = 3;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 0, 0);
        cameraPositionPanel.add(westLabel, gridBagConstraints);

        setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
        setTitle("Terrain Bender");
        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                formWindowClosing(evt);
            }
        });

        container3D.setOpaque(false);
        container3D.addComponentListener(new java.awt.event.ComponentAdapter() {
            public void componentResized(java.awt.event.ComponentEvent evt) {
                container3DComponentResized(evt);
            }
        });
        container3D.setLayout(new java.awt.BorderLayout());
        getContentPane().add(container3D, java.awt.BorderLayout.CENTER);

        eastPanel.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.CENTER, 0, 0));

        jPanel1.setLayout(new java.awt.GridBagLayout());

        settingsPanel.setLayout(new java.awt.GridBagLayout());

        gradationGraph.setFont(new java.awt.Font("Lucida Grande", 0, 10));
        gradationGraph.setLabel("Terrain base bending");
        gradationGraph.setShowDiagonal(false);
        gradationGraph.setShowGradientBars(false);
        gradationGraph.addPropertyChangeListener(new java.beans.PropertyChangeListener() {
            public void propertyChange(java.beans.PropertyChangeEvent evt) {
                gradationGraphPropertyChange(evt);
            }
        });
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.gridwidth = 3;
        settingsPanel.add(gradationGraph, gridBagConstraints);

        jLabel2.setFont(new java.awt.Font("Lucida Grande", 0, 10));
        jLabel2.setText("Front");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.SOUTHEAST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 4, 0);
        settingsPanel.add(jLabel2, gridBagConstraints);

        jLabel3.setFont(new java.awt.Font("Lucida Grande", 0, 10));
        jLabel3.setText("Back");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 4;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.SOUTHWEST;
        gridBagConstraints.insets = new java.awt.Insets(0, 2, 4, 5);
        settingsPanel.add(jLabel3, gridBagConstraints);

        bendingTabbedPane.addTab("Bending", settingsPanel);

        detailsPanel_.setLayout(new java.awt.GridBagLayout());

        jLabel1.setText("Background");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        detailsPanel_.add(jLabel1, gridBagConstraints);

        jLabel4.setText("Foreground");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        detailsPanel_.add(jLabel4, gridBagConstraints);

        scaleBackgroundSlider.setMajorTickSpacing(10);
        scaleBackgroundSlider.setPaintLabels(true);
        scaleBackgroundSlider.setPaintTicks(true);
        scaleBackgroundSlider.setValue(10);
        scaleBackgroundSlider.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                modelSliderStateChanged(evt);
            }
        });
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        detailsPanel_.add(scaleBackgroundSlider, gridBagConstraints);
        {
            java.util.Hashtable labels = new java.util.Hashtable();
            labels.put( new Integer( 0 ), new javax.swing.JLabel("0") );
            labels.put( new Integer( 10 ), new javax.swing.JLabel("1") );
            labels.put( new Integer( 50 ), new javax.swing.JLabel("5") );
            labels.put( new Integer( 100 ), new javax.swing.JLabel("10") );
            scaleBackgroundSlider.setLabelTable(labels);
        }

        scaleForegroundSlider.setMajorTickSpacing(10);
        scaleForegroundSlider.setPaintLabels(true);
        scaleForegroundSlider.setPaintTicks(true);
        scaleForegroundSlider.setValue(10);
        scaleForegroundSlider.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                modelSliderStateChanged(evt);
            }
        });
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 2;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        detailsPanel_.add(scaleForegroundSlider, gridBagConstraints);
        {
            java.util.Hashtable labels = new java.util.Hashtable();
            labels.put( new Integer( 0 ), new javax.swing.JLabel("0") );
            labels.put( new Integer( 10 ), new javax.swing.JLabel("1") );
            labels.put( new Integer( 50 ), new javax.swing.JLabel("5") );
            labels.put( new Integer( 100 ), new javax.swing.JLabel("10") );
            scaleForegroundSlider.setLabelTable(labels);
        }

        jLabel7.setText("Vertical exaggeration of the terrain");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridwidth = 2;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.insets = new java.awt.Insets(6, 0, 6, 0);
        detailsPanel_.add(jLabel7, gridBagConstraints);

        noVerticalExaggerationButton.setText("Reset Exaggeration");
        noVerticalExaggerationButton.setToolTipText("Reset the vertical exaggeration factors to 1 for the foreground and the background.");
        noVerticalExaggerationButton.setEnabled(false);
        noVerticalExaggerationButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                noVerticalExaggerationButtonActionPerformed(evt);
            }
        });
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 3;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        gridBagConstraints.insets = new java.awt.Insets(10, 0, 0, 0);
        detailsPanel_.add(noVerticalExaggerationButton, gridBagConstraints);

        detailsPanel.add(detailsPanel_);

        bendingTabbedPane.addTab("Vertical Exaggeration", detailsPanel);

        sphericalPanel_.setLayout(new java.awt.GridBagLayout());

        jLabel5.setText("Curvature");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.anchor = java.awt.GridBagConstraints.EAST;
        sphericalPanel_.add(jLabel5, gridBagConstraints);

        curvatureSlider.setMajorTickSpacing(45);
        curvatureSlider.setMaximum(180);
        curvatureSlider.setMinorTickSpacing(15);
        curvatureSlider.setPaintLabels(true);
        curvatureSlider.setPaintTicks(true);
        curvatureSlider.setValue(0);
        curvatureSlider.addChangeListener(new javax.swing.event.ChangeListener() {
            public void stateChanged(javax.swing.event.ChangeEvent evt) {
                modelSliderStateChanged(evt);
            }
        });
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        sphericalPanel_.add(curvatureSlider, gridBagConstraints);
        {
            java.util.Hashtable labels = curvatureSlider.createStandardLabels(45);
            java.util.Enumeration e = labels.elements();
            while(e.hasMoreElements()) {
                javax.swing.JComponent comp = (javax.swing.JComponent)e.nextElement();
                if (comp instanceof javax.swing.JLabel) {
                    javax.swing.JLabel label = (javax.swing.JLabel)(comp);
                    label.setText(label.getText() + "\u00b0");
                }
            }
            curvatureSlider.setLabelTable(labels);
        }

        horizonPanel.add(sphericalPanel_);

        bendingTabbedPane.addTab("Horizon", horizonPanel);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        jPanel1.add(bendingTabbedPane, gridBagConstraints);
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        jPanel1.add(map3DOptionsPanel, gridBagConstraints);

        directionButton.setText("Bend in Current View Direction");
        directionButton.setToolTipText("Set the direction of bending to the current view direction.");
        directionButton.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                directionButtonActionPerformed(evt);
            }
        });
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.insets = new java.awt.Insets(5, 0, 5, 0);
        jPanel1.add(directionButton, gridBagConstraints);
        directionButton.putClientProperty("JComponent.sizeVariant", "small");

        eastPanel.add(jPanel1);

        getContentPane().add(eastPanel, java.awt.BorderLayout.EAST);

        fileMenu.setText("File");
        fileMenu.addMenuListener(new javax.swing.event.MenuListener() {
            public void menuSelected(javax.swing.event.MenuEvent evt) {
                fileMenuMenuSelected(evt);
            }
            public void menuDeselected(javax.swing.event.MenuEvent evt) {
            }
            public void menuCanceled(javax.swing.event.MenuEvent evt) {
            }
        });

        openDEMMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_O,
            java.awt.Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
    openDEMMenuItem.setText("Open Terrain Model");
    openDEMMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            openDEMMenuItemActionPerformed(evt);
        }
    });
    fileMenu.add(openDEMMenuItem);

    openRecentMenu.setText("Open Recent Terrain Model");
    fileMenu.add(openRecentMenu);

    saveDEMMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_S,
        java.awt.Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
saveDEMMenuItem.setText("Save Bent Terrain Model");
saveDEMMenuItem.addActionListener(new java.awt.event.ActionListener() {
    public void actionPerformed(java.awt.event.ActionEvent evt) {
        saveDEMMenuItemActionPerformed(evt);
    }
    });
    fileMenu.add(saveDEMMenuItem);
    fileMenu.add(jSeparator5);

    openTextureMenuItem.setText("Open Texture Image");
    openTextureMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            openTextureMenuItemActionPerformed(evt);
        }
    });
    fileMenu.add(openTextureMenuItem);

    clearTextureMenuItem.setText("Remove Texture");
    clearTextureMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            clearTextureMenuItemActionPerformed(evt);
        }
    });
    fileMenu.add(clearTextureMenuItem);
    fileMenu.add(jSeparator1);

    saveSettingsMenuItem.setText("Save Settings");
    saveSettingsMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            saveSettingsMenuItemActionPerformed(evt);
        }
    });
    fileMenu.add(saveSettingsMenuItem);

    openSettingsMenuItem.setText("Open Settings");
    openSettingsMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            openSettingsMenuItemActionPerformed(evt);
        }
    });
    fileMenu.add(openSettingsMenuItem);
    fileMenu.add(jSeparator2);

    export3DViewMenuItem.setText("Export Rendering");
    export3DViewMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            export3DViewMenuItemActionPerformed(evt);
        }
    });
    fileMenu.add(export3DViewMenuItem);

    jMenu1.setText("Export Depth Buffer");

    exportDepthBuffer8BitMenuItem.setText("8 Bit");
    exportDepthBuffer8BitMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            exportDepthBuffer8BitMenuItemActionPerformed(evt);
        }
    });
    jMenu1.add(exportDepthBuffer8BitMenuItem);

    exportDepthBuffer16BitMenuItem.setText("16 Bit");
    exportDepthBuffer16BitMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            exportDepthBuffer16BitMenuItemActionPerformed(evt);
        }
    });
    jMenu1.add(exportDepthBuffer16BitMenuItem);

    fileMenu.add(jMenu1);
    fileMenu.add(exitMenuSeparator);

    exitMenuItem.setText("Exit");
    exitMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            exitMenuItemActionPerformed(evt);
        }
    });
    fileMenu.add(exitMenuItem);

    menuBar.add(fileMenu);

    editMenu.setText("Edit");
    editMenu.addMenuListener(new javax.swing.event.MenuListener() {
        public void menuSelected(javax.swing.event.MenuEvent evt) {
            editMenuMenuSelected(evt);
        }
        public void menuDeselected(javax.swing.event.MenuEvent evt) {
        }
        public void menuCanceled(javax.swing.event.MenuEvent evt) {
        }
    });

    cutMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_X,
        java.awt.Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
cutMenuItem.setText("Cut");
cutMenuItem.setEnabled(false);
editMenu.add(cutMenuItem);

copyMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_C,
    java.awt.Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
    copyMenuItem.setText("Copy");
    copyMenuItem.setEnabled(false);
    editMenu.add(copyMenuItem);

    pasteMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_V,
        java.awt.Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
pasteMenuItem.setText("Paste");
pasteMenuItem.setEnabled(false);
editMenu.add(pasteMenuItem);

deleteMenuItem.setText("Delete");
deleteMenuItem.setEnabled(false);
editMenu.add(deleteMenuItem);
editMenu.add(jSeparator3);

gridInfoMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_I,
    java.awt.Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
    gridInfoMenuItem.setText("Terrain Model Info");
    gridInfoMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            gridInfoMenuItemActionPerformed(evt);
        }
    });
    editMenu.add(gridInfoMenuItem);
    editMenu.add(jSeparator4);

    bendingDirectionMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_D,
        java.awt.Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
bendingDirectionMenuItem.setText("Bend in Current View Direction");
bendingDirectionMenuItem.addActionListener(new java.awt.event.ActionListener() {
    public void actionPerformed(java.awt.event.ActionEvent evt) {
        bendingDirectionMenuItemActionPerformed(evt);
    }
    });
    editMenu.add(bendingDirectionMenuItem);

    resetBendingMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_R,
        java.awt.Toolkit.getDefaultToolkit().getMenuShortcutKeyMask()));
resetBendingMenuItem.setText("Reset to Default Bending");
resetBendingMenuItem.addActionListener(new java.awt.event.ActionListener() {
    public void actionPerformed(java.awt.event.ActionEvent evt) {
        resetBendingMenuItemActionPerformed(evt);
    }
    });
    editMenu.add(resetBendingMenuItem);
    editMenu.add(jSeparator6);

    viewingDirectionMenuItem.setText("View in Bending Direction");
    viewingDirectionMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            viewingDirectionMenuItemActionPerformed(evt);
        }
    });
    editMenu.add(viewingDirectionMenuItem);
    editMenu.add(jSeparator7);

    scaleTerrainMenuItem.setText("Scale Terrain");
    scaleTerrainMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            scaleTerrainMenuItemActionPerformed(evt);
        }
    });
    editMenu.add(scaleTerrainMenuItem);
    editMenu.add(jSeparator8);

    cameraPositionMenuItem.setText("Position Cylindrical Camera");
    cameraPositionMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            cameraPositionMenuItemActionPerformed(evt);
        }
    });
    editMenu.add(cameraPositionMenuItem);

    menuBar.add(editMenu);

    macHelpMenu.setText("Help");

    macHelpMenuItem.setText("Terrain Bender Online Manual");
    macHelpMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            helpMenuItemActionPerformed(evt);
        }
    });
    macHelpMenu.add(macHelpMenuItem);

    systemInformationMenuItem.setText("System Information");
    systemInformationMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            systemInformationMenuItemActionPerformed(evt);
        }
    });
    macHelpMenu.add(systemInformationMenuItem);

    menuBar.add(macHelpMenu);

    winHelpMenu.setText("?");

    winHelpMenuItem.setText("Terrain Bender Online Manual");
    winHelpMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            helpMenuItemActionPerformed(evt);
        }
    });
    winHelpMenu.add(winHelpMenuItem);

    infoMenuItem.setText("Info");
    infoMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            infoMenuItemActionPerformed(evt);
        }
    });
    winHelpMenu.add(infoMenuItem);

    winSystemInformationMenuItem.setText("System Information");
    winSystemInformationMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            winSystemInformationMenuItemActionPerformed(evt);
        }
    });
    winHelpMenu.add(winSystemInformationMenuItem);

    menuBar.add(winHelpMenu);

    extrasMenu.setText("Extras");

    highResMenuItem.setText("Use High Resolution Cylindrical Rendering");
    highResMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            highResMenuItemActionPerformed(evt);
        }
    });
    extrasMenu.add(highResMenuItem);

    bindLightDirectionMenuItem.setText("Bind Light Direction to View Direction");
    bindLightDirectionMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            bindLightDirectionMenuItemActionPerformed(evt);
        }
    });
    extrasMenu.add(bindLightDirectionMenuItem);

    updateDisplayMenuItem.setText("Update Display");
    updateDisplayMenuItem.addActionListener(new java.awt.event.ActionListener() {
        public void actionPerformed(java.awt.event.ActionEvent evt) {
            updateDisplayMenuItemActionPerformed(evt);
        }
    });
    extrasMenu.add(updateDisplayMenuItem);

    menuBar.add(extrasMenu);
    menuBar.remove(extrasMenu);

    setJMenuBar(menuBar);

    pack();
    }// </editor-fold>//GEN-END:initComponents

    private void resetBaseCurve() {
        p.curve.fromString("0 0 0.25 0.25 1 0.5");
        this.gradationGraph.setCurve(p.curve); // fires event
    }

    /**
     * Applies the bending to the full resolution grid and writes it to an
     * ESRI ASCII file.
     * @param filePath
     */
    private void writeBentDEM(final String filePath) {

        final JFrame frame = this;
        final boolean cylindrical = this.map3DViewer.getCamera() == Camera.cylindrical;

        SwingWorkerWithProgressIndicator worker;
        worker = new SwingWorkerWithProgressIndicator(
                this, "Terrain Bender Export", "", true) {

            @Override
            public void done() {
                // check for exceptions
                try {
                    get();
                } catch (Exception ex) {
                    // clean up the file
                    new File(filePath).delete();

                    // show error message if the exception was not caused by the user
                    String exmsg = ex.getMessage();
                    if (exmsg != null && exmsg.contains("user canceled")) {
                        return;
                    }

                    String msg = "An error occured. The file could not be generated.";
                    String title = "Terrain Bender Error";
                    ika.utils.ErrorDialog.showErrorDialog(msg, title, ex, frame);
                    ex.printStackTrace();
                }
            }

            @Override
            protected Object doInBackground() throws Exception {
                PrintWriter pw = new PrintWriter(filePath, "US-ASCII");
                try {
                    // open progress dialog
                    this.start();
                    
                    // first bend the grid to find the smallest and largest
                    // values in the bent grid to determine a
                    // value for void cells.
                    MinMaxFinderGrid minMaxFinder = new MinMaxFinderGrid(null);
                    bendGrid(getFullResolutionGrid(), minMaxFinder, cylindrical);
                    
                    // bend and write the grid
                    float min = minMaxFinder.getMinValue();
                    float max = minMaxFinder.getMaxValue();

                    // vertically shift the terrain to set the smallest value to 0
                    float offset = -min;

                    // if the vertically shifted maximum value is larger than
                    // what Natural Scene Designer supports, further shift the terrain
                    if (max - min > NSD_MAX) {
                        offset = -min + NSD_MIN;
                    }
                    ESRIASCIIGridWriterWrapper writer;
                    writer = new ESRIASCIIGridWriterWrapper(pw, this, offset);
                    writer.init(getFullResolutionGrid());
                    bendGrid(getFullResolutionGrid(), writer, cylindrical);
                    return null;
                } finally {
                    if (pw != null) {
                        pw.close();
                    }
                    this.complete();
                }

            }
        };

        worker.setMaxTimeWithoutDialog(1);
        worker.setMessage("Writing Terrain Model \"" + FileUtils.getFileName(filePath) + "\"");
        worker.execute();
    }

private void saveDEMMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_saveDEMMenuItemActionPerformed

    if (getFullResolutionGrid() == null) {
        noTerrainErrorMessage(this);
        return;
    }

    String name = this.getTitle();
    name = FileUtils.cutFileExtension(name);
    if (name.isEmpty()) {
        name = "bent_grid.asc";
    } else {
        name += name.contains(" ") ? " bent.asc" : "_bent.asc";
    }
    String path = FileUtils.askFile(null, "Export ESRI ASCII Grid", name, false, "asc");
    if (path == null) {
        return;
    }
    this.writeBentDEM(path);
}//GEN-LAST:event_saveDEMMenuItemActionPerformed

private void saveSettingsMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_saveSettingsMenuItemActionPerformed

    try {
        String filePath = FileUtils.askFile(this, "Save Settings", "Settings.txt", false, "txt");
        if (filePath == null) {
            return;
        }
        String serializedModel = this.toString();
        BufferedWriter out = new BufferedWriter(new FileWriter(filePath));
        out.write(serializedModel);
        out.close();
    } catch (IOException e) {
        ika.utils.ErrorDialog.showErrorDialog("The settings could not be written.", e);
    }

}//GEN-LAST:event_saveSettingsMenuItemActionPerformed

private void openSettingsMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_openSettingsMenuItemActionPerformed
    try {
        String newFilePath = FileUtils.askFile(this, "Open Settings", true);
        if (newFilePath == null) {
            return; // user canceled
        }

        // import data from the file
        File file = new File(newFilePath);
        byte[] data = FileUtils.getBytesFromFile(file);
        this.fromString(new String(data));
        this.modelChanged();

    } catch (IOException e) {
        String msg = "The settings could not be read.";
        String title = "Terrain Bender Error";
        ika.utils.ErrorDialog.showErrorDialog(msg, title, e, this);
    }
}//GEN-LAST:event_openSettingsMenuItemActionPerformed

private void exportImage(String fileDialogInfo, boolean zBuffer, boolean bit16) {
    try {
        String path = FileUtils.askFile(null, fileDialogInfo, "3d.png", false, "png");
        if (path != null) {
            BufferedImage image;
            image = zBuffer ? map3DViewer.getDepthBufferImage(bit16) : map3DViewer.getImage();
            ImageIO.write(image, "png", new File(path));
        }
    } catch (IOException ex) {
        String msg = "The image could not be saved.";
        String title = "Terrain Bender Error";
        ika.utils.ErrorDialog.showErrorDialog(msg, title, ex, this);
    }
}
private void export3DViewMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_export3DViewMenuItemActionPerformed
    exportImage("Save 3D View", false, false);
}//GEN-LAST:event_export3DViewMenuItemActionPerformed

private void container3DComponentResized(java.awt.event.ComponentEvent evt) {//GEN-FIRST:event_container3DComponentResized
    if (map3DViewer != null) {
        this.map3DViewer.display();
    }
}//GEN-LAST:event_container3DComponentResized

private void gridInfoMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_gridInfoMenuItemActionPerformed
    showGridInfo();
}//GEN-LAST:event_gridInfoMenuItemActionPerformed

private void formWindowClosing(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowClosing
    askToCloseWindowAndExit();
}//GEN-LAST:event_formWindowClosing

private void bendingDirectionMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bendingDirectionMenuItemActionPerformed
    p.directionAngle = this.map3DViewer.getZAngle();
    modelChanged();
}//GEN-LAST:event_bendingDirectionMenuItemActionPerformed

private void infoMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_infoMenuItemActionPerformed
    ika.gui.ProgramInfoPanel.showApplicationInfo();
}//GEN-LAST:event_infoMenuItemActionPerformed

private void exitMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exitMenuItemActionPerformed
    askToCloseWindowAndExit();
}//GEN-LAST:event_exitMenuItemActionPerformed

private void gradationGraphPropertyChange(java.beans.PropertyChangeEvent evt) {//GEN-FIRST:event_gradationGraphPropertyChange
    if (GradationGraph.CURVE_CHANGED.equals(evt.getPropertyName())) {
        modelChanged();
    }
}//GEN-LAST:event_gradationGraphPropertyChange

private void modelSliderStateChanged(javax.swing.event.ChangeEvent evt) {//GEN-FIRST:event_modelSliderStateChanged
    if (!((JSlider) evt.getSource()).getValueIsAdjusting()) {
        p.horizonAngle = curvatureSlider.getValue();
        p.backgroundScale = scaleBackgroundSlider.getValue() / 10f;
        p.foregroundScale = scaleForegroundSlider.getValue() / 10f;
        p.shiftX = map3DViewer.getShiftX();
        p.shiftY = map3DViewer.getShiftY();
        
        // update the enabled state of the button for resetting the vertical 
        // exaggeration factor
        boolean enableReset = p.backgroundScale != 1 || p.foregroundScale != 1;
        noVerticalExaggerationButton.setEnabled(enableReset);
        
        this.modelChanged();

    }
}//GEN-LAST:event_modelSliderStateChanged

private void directionButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_directionButtonActionPerformed
    p.directionAngle = this.map3DViewer.getZAngle();
    this.modelChanged();
}//GEN-LAST:event_directionButtonActionPerformed

private void openDEMMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_openDEMMenuItemActionPerformed
    this.openDEM();
}//GEN-LAST:event_openDEMMenuItemActionPerformed

private void openTextureMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_openTextureMenuItemActionPerformed
    String path = FileUtils.askFile(this, "Load Texture", true);
    if (path == null) {
        return;
    }
    openTexture(path);
}//GEN-LAST:event_openTextureMenuItemActionPerformed

private void clearTextureMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_clearTextureMenuItemActionPerformed
    this.map3DViewer.clearTextureImage();
}//GEN-LAST:event_clearTextureMenuItemActionPerformed

private void viewingDirectionMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_viewingDirectionMenuItemActionPerformed
    this.map3DViewer.setZAngle(this.p.directionAngle);
}//GEN-LAST:event_viewingDirectionMenuItemActionPerformed

private void resetBendingMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_resetBendingMenuItemActionPerformed
    this.p.directionAngle = Model.DEFAULT_BENDING_DIRECTION;
    resetBaseCurve();
}//GEN-LAST:event_resetBendingMenuItemActionPerformed

private void helpMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_helpMenuItemActionPerformed
    Properties props =
            ika.utils.PropertiesLoader.loadProperties("ika.app.Application.properties");
    String url = props.getProperty("HelpWebPage");
    if (Desktop.isDesktopSupported()) {
            try {
                URI uri = new URI(url);
                Desktop.getDesktop().browse(uri);
            } catch (URISyntaxException ex) {
                Logger.getLogger(TerrainBenderWindow.class.getName()).log(Level.SEVERE, null, ex);
            } catch (IOException ex) {
                Logger.getLogger(TerrainBenderWindow.class.getName()).log(Level.SEVERE, null, ex);
            }
    }
    //ika.utils.BrowserLauncherWrapper.openURL(url);
}//GEN-LAST:event_helpMenuItemActionPerformed

private void scaleTerrainMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_scaleTerrainMenuItemActionPerformed

    if (getFullResolutionGrid() == null) {
        noTerrainErrorMessage(this);
        return;
    }

    int option = JOptionPane.showOptionDialog(this,
            scaleTerrainPanel,
            "Scale Terrain",
            JOptionPane.OK_CANCEL_OPTION,
            JOptionPane.QUESTION_MESSAGE,
            null, null, null);
    if (option != JOptionPane.OK_OPTION) {
        return;
    }

    try {
        scaleTerrainFormattedTextField.commitEdit();
        java.lang.Number f = (java.lang.Number) (scaleTerrainFormattedTextField.getValue());
        float scale = f.floatValue();
        GridScaleOperator op = new GridScaleOperator(scale);
        GeoGrid scaledGrid = op.operate(getFullResolutionGrid());
        gridPyramid = new GaussianPyramid(scaledGrid, 9999, MIN_GRID_CELLS);
        originalInfoText = getFullResolutionGrid().toStringWithStatistics("<br>");
        modelChanged();
    } catch (Exception exc) {
        ErrorDialog.showErrorDialog("An error occured while scaling the terrain.",
                "Error", exc, this);
    }

}//GEN-LAST:event_scaleTerrainMenuItemActionPerformed

private void cameraPositionMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_cameraPositionMenuItemActionPerformed

    // get grid dimensions
    GeoGrid grid = this.getFullResolutionGrid();

    if (grid == null) {
        noTerrainErrorMessage(this);
        return;
    }

    double w = grid.getBounds2D(GeoObject.UNDEFINED_SCALE).getWidth();
    double h = grid.getBounds2D(GeoObject.UNDEFINED_SCALE).getHeight();
    double cx = grid.getCenterX(GeoObject.UNDEFINED_SCALE);
    double cy = grid.getCenterY(GeoObject.UNDEFINED_SCALE);

    // init extension labels
    DecimalFormat format = new DecimalFormat("#,##0.###");
    westLabel.setText(format.format(grid.getWest()));
    eastLabel.setText(format.format(grid.getEast()));
    southLabel.setText(format.format(grid.getSouth()));
    northLabel.setText(format.format(grid.getNorth()));

    // write current coordinates
    double xrel = this.map3DViewer.getShiftX();
    double yrel = this.map3DViewer.getShiftY();
    double x = xrel * w / 2 + cx;
    double y = yrel * h / 2 + cy;
    cameraXFormattedTextField.setValue(x);
    cameraYFormattedTextField.setValue(y);

    // show dialog
    int option = JOptionPane.showOptionDialog(this,
            cameraPositionPanel,
            "Position Cylindrical Camera",
            JOptionPane.OK_CANCEL_OPTION,
            JOptionPane.QUESTION_MESSAGE,
            null, null, null);

    if (option != JOptionPane.OK_OPTION) {
        return;
    }

    // read new coordinates
    try {
        cameraXFormattedTextField.commitEdit();
        cameraYFormattedTextField.commitEdit();

        x = ((Number) (cameraXFormattedTextField.getValue())).doubleValue();
        y = ((Number) (cameraYFormattedTextField.getValue())).doubleValue();

        this.map3DViewer.setCamera(Camera.cylindrical);

        xrel = (x - cx) / w * 2;
        yrel = (y - cy) / h * 2;

        this.map3DViewer.setShiftX((float) xrel);
        this.map3DViewer.setShiftY((float) yrel);

    } catch (Exception exc) {
        ErrorDialog.showErrorDialog("An error occured while positioning the camera.",
                "Error", exc, this);
    }
}//GEN-LAST:event_cameraPositionMenuItemActionPerformed

private void noVerticalExaggerationButtonActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_noVerticalExaggerationButtonActionPerformed
    if (scaleBackgroundSlider.getValue() != 10 || scaleForegroundSlider.getValue() != 10) {
        scaleBackgroundSlider.setValue(10);
        scaleForegroundSlider.setValue(10);
    }
}//GEN-LAST:event_noVerticalExaggerationButtonActionPerformed

private void highResMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_highResMenuItemActionPerformed
    map3DViewer.setHighResCylindricalRendering(highResMenuItem.isSelected());
}//GEN-LAST:event_highResMenuItemActionPerformed

private void bindLightDirectionMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_bindLightDirectionMenuItemActionPerformed
    map3DViewer.setBindLightDirectionToViewDirection(bindLightDirectionMenuItem.isSelected());
}//GEN-LAST:event_bindLightDirectionMenuItemActionPerformed

private void winSystemInformationMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_winSystemInformationMenuItemActionPerformed
    new SystemInfoGPU(this);
}//GEN-LAST:event_winSystemInformationMenuItemActionPerformed

private void systemInformationMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_systemInformationMenuItemActionPerformed
    new SystemInfoGPU(this);
}//GEN-LAST:event_systemInformationMenuItemActionPerformed

private void updateDisplayMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_updateDisplayMenuItemActionPerformed
    map3DViewer.display();
}//GEN-LAST:event_updateDisplayMenuItemActionPerformed

private void exportDepthBuffer8BitMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exportDepthBuffer8BitMenuItemActionPerformed
    exportImage("Save Depth Buffer (8 Bits)", true, false);
}//GEN-LAST:event_exportDepthBuffer8BitMenuItemActionPerformed

private void exportDepthBuffer16BitMenuItemActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_exportDepthBuffer16BitMenuItemActionPerformed
    exportImage("Save Depth Buffer (16 Bits)", true, true);
}//GEN-LAST:event_exportDepthBuffer16BitMenuItemActionPerformed

private void fileMenuMenuSelected(javax.swing.event.MenuEvent evt) {//GEN-FIRST:event_fileMenuMenuSelected
    openTextureMenuItem.setEnabled(getFullResolutionGrid() != null);
    clearTextureMenuItem.setEnabled(map3DViewer.isTextureEnabledAndLoaded());
    if (map3DViewer.isTextureEnabledAndLoaded()) {
        if (map3DViewer.isTexture1D()) {
            clearTextureMenuItem.setText("Remove Hypsometric Tints");
        } else {
            clearTextureMenuItem.setText("Remove Texture");
        }
    }
}//GEN-LAST:event_fileMenuMenuSelected

private void editMenuMenuSelected(javax.swing.event.MenuEvent evt) {//GEN-FIRST:event_editMenuMenuSelected
    boolean hasTerrain = getFullResolutionGrid() != null;
    gridInfoMenuItem.setEnabled(hasTerrain);
    bendingDirectionMenuItem.setEnabled(hasTerrain);
    viewingDirectionMenuItem.setEnabled(hasTerrain);
    scaleTerrainMenuItem.setEnabled(hasTerrain);
    cameraPositionMenuItem.setEnabled(hasTerrain);

}//GEN-LAST:event_editMenuMenuSelected

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JMenuItem bendingDirectionMenuItem;
    private javax.swing.JTabbedPane bendingTabbedPane;
    private javax.swing.JCheckBoxMenuItem bindLightDirectionMenuItem;
    private javax.swing.JMenuItem cameraPositionMenuItem;
    private javax.swing.JPanel cameraPositionPanel;
    private javax.swing.JFormattedTextField cameraXFormattedTextField;
    private javax.swing.JFormattedTextField cameraYFormattedTextField;
    private javax.swing.JMenuItem clearTextureMenuItem;
    private javax.swing.JPanel container3D;
    private javax.swing.JMenuItem copyMenuItem;
    private javax.swing.JSlider curvatureSlider;
    private javax.swing.JMenuItem cutMenuItem;
    private javax.swing.JMenuItem deleteMenuItem;
    private javax.swing.JButton directionButton;
    private javax.swing.JLabel eastLabel;
    private javax.swing.JPanel eastPanel;
    private javax.swing.JMenu editMenu;
    private javax.swing.JMenuItem exitMenuItem;
    private javax.swing.JSeparator exitMenuSeparator;
    private javax.swing.JMenuItem export3DViewMenuItem;
    private javax.swing.JMenuItem exportDepthBuffer16BitMenuItem;
    private javax.swing.JMenuItem exportDepthBuffer8BitMenuItem;
    private javax.swing.JMenu extrasMenu;
    private javax.swing.JMenu fileMenu;
    private ika.gui.GradationGraph gradationGraph;
    private javax.swing.JMenuItem gridInfoMenuItem;
    private javax.swing.JCheckBoxMenuItem highResMenuItem;
    private javax.swing.JPanel horizonPanel;
    private javax.swing.JMenuItem infoMenuItem;
    private javax.swing.JLabel jLabel15;
    private javax.swing.JLabel jLabel16;
    private javax.swing.JLabel jLabel18;
    private javax.swing.JLabel jLabel20;
    private javax.swing.JLabel jLabel7;
    private javax.swing.JMenu jMenu1;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JSeparator jSeparator1;
    private javax.swing.JSeparator jSeparator2;
    private javax.swing.JSeparator jSeparator3;
    private javax.swing.JSeparator jSeparator4;
    private javax.swing.JSeparator jSeparator5;
    private javax.swing.JSeparator jSeparator6;
    private javax.swing.JSeparator jSeparator7;
    private javax.swing.JSeparator jSeparator8;
    private javax.swing.JMenu macHelpMenu;
    private javax.swing.JMenuItem macHelpMenuItem;
    private ch.ethz.karto.map3d.gui.Map3DOptionsPanel map3DOptionsPanel;
    private javax.swing.JMenuBar menuBar;
    private javax.swing.JButton noVerticalExaggerationButton;
    private javax.swing.JLabel northLabel;
    private javax.swing.JMenuItem openDEMMenuItem;
    private javax.swing.JMenu openRecentMenu;
    private javax.swing.JMenuItem openSettingsMenuItem;
    private javax.swing.JMenuItem openTextureMenuItem;
    private javax.swing.JMenuItem pasteMenuItem;
    private javax.swing.JMenuItem resetBendingMenuItem;
    private javax.swing.JMenuItem saveDEMMenuItem;
    private javax.swing.JMenuItem saveSettingsMenuItem;
    private javax.swing.JSlider scaleBackgroundSlider;
    private javax.swing.JSlider scaleForegroundSlider;
    private javax.swing.JFormattedTextField scaleTerrainFormattedTextField;
    private javax.swing.JMenuItem scaleTerrainMenuItem;
    private javax.swing.JPanel scaleTerrainPanel;
    private javax.swing.JLabel southLabel;
    private javax.swing.JMenuItem systemInformationMenuItem;
    private javax.swing.JMenuItem updateDisplayMenuItem;
    private javax.swing.JMenuItem viewingDirectionMenuItem;
    private javax.swing.JLabel westLabel;
    private javax.swing.JMenu winHelpMenu;
    private javax.swing.JMenuItem winHelpMenuItem;
    private javax.swing.JMenuItem winSystemInformationMenuItem;
    // End of variables declaration//GEN-END:variables

    @Override
    public void propertyChange(PropertyChangeEvent evt) {

        if ("camera changed".equals(evt.getPropertyName())) {
            // camera changed: generate new bent sourceGrid for the new camera
            modelChanged();

            // disable spherical curvature for cylindrical projection
            boolean cylindricalCamera = this.map3DViewer.getCamera() == Camera.cylindrical;
            // disable the sphere tab
            bendingTabbedPane.setEnabledAt(2, !cylindricalCamera);
            // disable controls in sphere tap
            curvatureSlider.setEnabled(!cylindricalCamera);

            // disable the button and menu item for adjusting the direction of bending
            directionButton.setEnabled(!cylindricalCamera);
            bendingDirectionMenuItem.setEnabled(!cylindricalCamera);
            viewingDirectionMenuItem.setEnabled(!cylindricalCamera);

        } else if ("shiftX".equals(evt.getPropertyName())) {
            p.shiftX = map3DViewer.getShiftX();
            if (map3DViewer.getCamera() == Camera.cylindrical) {
                modelChanged();
            }
        } else if ("shiftY".equals(evt.getPropertyName())) {
            p.shiftY = map3DViewer.getShiftY();
            if (map3DViewer.getCamera() == Camera.cylindrical) {
                modelChanged();
            }
        }
    }

    /**
     * Part of the ComponentListener interface.
     */
    @Override
    public void componentShown(ComponentEvent e) {
    }

    /**
     * Part of the ComponentListener interface.
     */
    @Override
    public void componentHidden(ComponentEvent e) {
    }

    /**
     * Part of the ComponentListener interface.
     */
    @Override
    public void componentMoved(ComponentEvent e) {
    }

    /**
     * Part of the ComponentListener interface. Make sure this window is not
     * getting too small.
     */
    @Override
    public void componentResized(ComponentEvent e) {

        if (this.packedSize == null) {
            return;
        }

        // Check if either the width or the height are below minimum
        // and reset size if necessary.
        // Note: this is not elegant, but SUN recommends doing it that way.
        int width = getWidth();
        int height = getHeight();
        boolean resize = false;

        if (width < this.packedSize.width) {
            resize = true;
            width = this.packedSize.width;
        }
        if (height < this.packedSize.height) {
            resize = true;
            height = this.packedSize.height;
        }
        if (resize) {
            setSize(width, height);
        }
    }
}
